from xmlrpc.client import DateTime
from .frigel_report_creator import FrigelReportCreator
from .consts import *
from .translations import Translations
from .charts.charts import ChartPieWithLegend, ChartBar, PieChartWorkingModes

from reportlab.lib.pagesizes import A4
from reportlab.lib.styles import getSampleStyleSheet
from reportlab.lib.units import mm, cm
from reportlab.graphics.shapes import Circle, Drawing, _DrawingEditorMixin
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.charts.barcharts import VerticalBarChart
from reportlab.graphics.charts.piecharts import Pie
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas
from reportlab.platypus import Image, Paragraph, Table, TableStyle, KeepInFrame, Frame
from reportlab.graphics.charts.textlabels import Label
from reportlab.lib.colors import HexColor
from reportlab.lib import colors
from reportlab.lib.colors import PCMYKColor
from reportlab.lib.formatters import DecimalFormatter

from typing import Union
from io import BytesIO
from datetime import datetime, timedelta
import numpy as np

class SyncroReportWeeklyCreator(FrigelReportCreator):
    '''
    Class for the creation of Syncro RS reports
    '''
    
    def __init__(self, operation_hours: list, operation_hours_prev: list, working_modes_hours: dict, working_modes_hours_prev: dict, start_date: datetime):
        '''
        input:
            operation_hours (list): list with the operation hours of the week
            operation_hours_prev (list): list with the operation hours of the previous week
            working_modes_hours (dict): {'standard': [10, 20,..], 'production': [10, 20,..], 'maintenance': [10, 20,..]}
            start_date (datetime): starting day
        '''
        self.operation_hours = operation_hours
        self.total_operation_hours = np.sum([0 if x == None else x for x in self.operation_hours])
        self.operation_hours_prev = operation_hours_prev
        self.total_operation_hours_prev = None if all(x is None for x in self.operation_hours_prev) else np.sum([0 if x == None else x for x in self.operation_hours_prev])
        self.working_modes_hours = working_modes_hours
        self.working_modes_hours_prev = working_modes_hours_prev
        self.working_modes_trend = [np.sum([0 if x == None else x for x in self.working_modes_hours[mode]])/np.sum([0 if x == None else x for x in self.working_modes_hours_prev[mode]])*100-100 if np.sum(0 if x == None else x for x in self.working_modes_hours_prev[mode])>0 else None\
            for mode in WORKING_MODE_TYPES]
        self.start_date = start_date
        FrigelReportCreator.__init__(self)
        
           
    def generatePDF(self, dest_path=None, language: str='en') -> Union[BytesIO, str]:
        '''
        generate PDF file
        
        input:
            dest_path (str): save path, if None the PDF will be saved into a ByetsIO buffer
            
        output:
            dest_path if is not None, else the buffer containing the PDF
        '''
        if dest_path:
            pdf_file = dest_path
        else:
            pdf_file = BytesIO()
    
        self.canvas = canvas.Canvas(pdf_file , pagesize=A4)
        self.styles = getSampleStyleSheet()
        self.width, self.height = A4
        self.organization = 'Frigel Firenze S.p.A.'
        self.logoPositionX = 5
        self.logoPositionY = 0
        self.horizontalMargin = 10
        self.lastPositionX = 0
        self.lastPositionY = DEFAULT_STARTING_POSITION
        
        self.translations = Translations(language)
        
        self.drawLogo()
        
        self.__add_machine_operation_paragraph()
        self.__add_operation_hours_pie()
        
        self.__add_working_hours_paragraph()
        self.__add_working_hours_chart()
        
        
        self.newPage()
        
        self.__add_working_modes_paragraph()
        self.__add_working_modes_data()
        
        self.__add_modes_hours_paragraph()
        self.__add_modes_hours_chart()
        
        self.canvas.save()
        
        return pdf_file
    
    # helper functions to create syncro PDF report
    #----------------------------------------------------------------------
    def __add_machine_operation_paragraph(self):
        text_machine = """<font size="24">
        {}
        </font>
        """.format(self.translations.getTranslation('Machine operation'))
        self.addSpacing(DEFAULT_SPACING)
        p_machine = Paragraph(text_machine, self.styles["Heading1"])
        self.drawOnCanvas(p_machine, 0, 0)
        
        
    def __add_operation_hours_pie(self):
        pie_table_width = self.width/2-self.horizontalMargin - 20
        drawing_circle = Drawing(CIRCLE_SIZE,CIRCLE_SIZE)
        drawing_circle.add(Circle(CIRCLE_SIZE,CIRCLE_SIZE,CIRCLE_SIZE, fillColor=DEFAULT_ON_COLOR, strokeColor=DEFAULT_STROKE_COLOR))
        
        table_on_time = Table(data=[[drawing_circle, self.translations.getTranslation('ON Time')]], colWidths=[CIRCLE_SIZE+8, None])
        table_on_time.setStyle([("FONTSIZE", (1,0), (1,0), 14),
                                ("FONTNAME", (1,0), (1,0), DEFAULT_FONT_BOLD),
                                ("VALIGN", (0,0), (-1,-1), 'BOTTOM')])
        table_on_hours = Table(data=[['{} h'.format(int(self.total_operation_hours))]])
        table_on_hours.setStyle([("FONTSIZE", (0,0), (0,0), 42),
                                 ("FONTSIZE", (1,0), (1,0), 14),
                                 ("FONTNAME", (0,0), (-1,-1), DEFAULT_FONT),
                                 ("VALIGN", (0,0), (-1,-1), 'BOTTOM')])
        
        data_table_right = [[table_on_time],
                            [table_on_hours],
                            ['{}%'.format(int((self.total_operation_hours/self.total_operation_hours_prev)*100 if self.total_operation_hours_prev > 0 else 0))\
                                if self.total_operation_hours_prev != None else ''],
                            [self.translations.getTranslation('vs {}h (prev 7 days)').format(self.total_operation_hours_prev)] if self.total_operation_hours_prev != None else '']
        table_pie_right = Table(data_table_right, colWidths=pie_table_width/2, rowHeights=[None, None, 50, None])
        table_pie_right.setStyle([("ALIGN", (0,0), (-1,-1), "LEFT"),
                                  ("FONTSIZE", (0,0), (0,0), 14),
                                  ("FONTSIZE", (0,1), (0,1),  42),
                                  ("FONTNAME", (0,0), (-1,-1), DEFAULT_FONT),
                                  ("FONTSIZE", (0,2), (0,2), 14)])
        
        hours_on_percentage = round(self.total_operation_hours*100/(24*7), 1)
        pie_data = [hours_on_percentage, 100-hours_on_percentage]
        pie_chart_hours = ChartPieWithLegend(data=pie_data, seriesNames=[self.translations.getTranslation('ON'), self.translations.getTranslation('OFF')], _colors=[DEFAULT_ON_COLOR, DEFAULT_OFF_COLOR])
        
        data_pie_hours = [[pie_chart_hours, table_pie_right]]
        table_pie_hours = Table(data_pie_hours, colWidths=pie_table_width)
        table_pie_hours.setStyle([("VALIGN", (0,0), (-1,-1), "MIDDLE"),
                            ("ALIGN", (1,0), (1,0), "CENTRE")])
        _, table_height = table_pie_hours.wrap(0, 0)
        self.addSpacing(DEFAULT_SPACING)
        self.drawOnCanvas(table_pie_hours, 0, table_height/3)
        
        
    def __add_working_hours_paragraph(self):
        text_machine = """<font size="24">
        {}
        </font>
        """.format(self.translations.getTranslation('Working hours (last 7 days)'))
        self.addSpacing(-15)
        p_machine = Paragraph(text_machine, self.styles["Heading1"])
        self.drawOnCanvas(p_machine, 0, 0)
        
        
    def __add_working_hours_chart(self):
        chart_data = [tuple(self.operation_hours), tuple(24-x if x != None else 0 for x in self.operation_hours)]
        chart_working_hours = ChartBar(data=chart_data,
                                       category_names=[(self.start_date+timedelta(days=x)).strftime("%d / %m") for x in range(7)],
                                       colors=[DEFAULT_ON_COLOR, DEFAULT_OFF_COLOR],
                                       style='stacked')
        self.addSpacing(DEFAULT_SPACING)
        self.drawOnCanvas(chart_working_hours, 0, chart_working_hours.height/3)
        
        
    def __add_working_modes_paragraph(self):
        text_machine = """<font size="24">
        {}
        </font>
        """.format(self.translations.getTranslation('Working modes'))
        self.addSpacing(DEFAULT_SPACING)
        p_machine = Paragraph(text_machine, self.styles["Heading1"])
        self.drawOnCanvas(p_machine, 0, 0)
        
        
    def __add_working_modes_data(self):
        data_pie = [np.sum(self.working_modes_hours[mode]) for mode in WORKING_MODE_TYPES]
        pie_working_modes = PieChartWorkingModes(data=data_pie,
                                                 data_trend=self.working_modes_trend,
                                                 header_names=(self.translations.getTranslation('Phases'), 
                                                               self.translations.getTranslation('Time'),
                                                               self.translations.getTranslation('Trend (prev. week)'),),
                                                 series_names=[self.translations.getTranslation('Standard'),
                                                               self.translations.getTranslation('Production'),
                                                               self.translations.getTranslation('Maintenance')],)
        self.drawOnCanvas(pie_working_modes, 0, pie_working_modes.height/3)
        
    def __add_modes_hours_paragraph(self):
        text_machine = """<font size="24">
        {}
        </font>
        """.format(self.translations.getTranslation('Phases hours (last 7 days)'))
        p_machine = Paragraph(text_machine, self.styles["Heading1"])
        self.addSpacing(-10)
        self.drawOnCanvas(p_machine, 0, 0)
        
    def __add_modes_hours_chart(self):
        chart_data = [self.working_modes_hours[mode] for mode in WORKING_MODE_TYPES]
        chart_working_hours = ChartBar(data=chart_data,
                                       category_names=[(self.start_date+timedelta(days=x)).strftime("%d / %m") for x in range(7)],
                                       colors=WORKING_MODES_COLORS,)
        
        self.addSpacing(DEFAULT_SPACING)
        self.drawOnCanvas(chart_working_hours, 0, chart_working_hours.height/3)
    #----------------------------------------------------------------------


from .consts import *

from reportlab.platypus import Image, Paragraph
from reportlab.lib.utils import ImageReader
from reportlab.lib.units import mm, cm

from abc import abstractclassmethod
from typing import Union
from io import BytesIO


class FrigelReportCreator():
    '''
    Base class for the creation of machine reports
    '''
    
    def __init__(self):
        self.logo = DEFAULT_LOGO_PATH


    @abstractclassmethod
    def generatePDF(self, dest_path=None, language: str='en') -> Union[BytesIO, str]:
        '''
        generate PDF file
        
        input:
            dest_path (str): save path, if None the PDF will be saved into a ByetsIO buffer
            language (str): translation language
            
        output:
            dest_path if is not None, else the buffer containing the PDF
        '''
        pass
    

    def getPDFAsBuffer(self, language: str='en') -> BytesIO:
        '''
        returns a buffer containing the PDF
        
        input:
            language (str): translation language
        
        output:
            buffer (BytesIO): buffer containing the PDF
        '''
        return self.generatePDF(language=language)
    
    
    def savePDF(self, dest_path: str, language: str='en') -> str:
        '''
        generates and saves the PDF to a file
        
        input:
            dest_path (str): save path
            language (str): translation language
            
        output:
            dest_path (str)
        '''
        return self.generatePDF(dest_path=dest_path, language=language)
    
    
    def get_image(self, path: str, width: float=1*cm) -> Image:
        '''
        returns image object mantaining aspect ratio
        
        input:
            path (str): image path
            width (float): expected width of the object
            
        output:
            image (reportlab.platypusImage)
        '''
        
        image = ImageReader(path)
        iw, ih = image.getSize()
        aspect = ih / float(iw)
        return Image(path, width=width, height=(width * aspect))
    
    
    def addSpacing(self, spacing: float) -> None:
        '''
        adds vertical spacing to the PDF
        
        input:
            spacing (float)
        '''
        self.lastPositionY += spacing
        
        
    def newPage(self) -> None:
        '''
        creates a new pdf page
        '''
        self.canvas.showPage()
        self.lastPositionX = 0
        self.lastPositionY = DEFAULT_STARTING_POSITION
        self.drawLogo()
        
    
    def drawOnCanvas(self, object: any, x: float, y: float) -> None:
        '''
        draws an object on canvas based on current position
        
        input:
            object (any): any reportlab object
            x (float): x offset
            y (float): y offset
        '''
        width, height = object.wrapOn(self.canvas, self.width, self.height)
        object.drawOn(self.canvas, *self.coord(self.horizontalMargin + x, self.lastPositionY + y, mm))
        self.lastPositionX = width/mm + self.horizontalMargin + x
        self.lastPositionY += height/mm + y
        
        
    def drawLogo(self) -> None:
        '''
        draws the logo on the page
        '''
        logo_object = self.get_image(self.logo, 3*cm)
        logo_object.wrapOn(self.canvas, self.width, self.height)
        logo_object.drawOn(self.canvas, *self.coord(self.logoPositionX, self.logoPositionY + logo_object.drawHeight/2, mm))
        
        
    def coord(self, x, y, unit=1) -> None:
        '''
        returns the coordinates relative to the pdf file
        
        input:
            x (float): x
            y (float): y
            
        output:
            x, y (float, float)
        '''
        x, y = x * unit, self.height -  y * unit
        return x, y    
    
    
    def createParagraph(self, text, x: float, y: float, style: any=None) -> None:
        '''
        add a paragraph to the PDF file
        
        input:
            text (str): paragraph text
            x (float): x position
            y (float): y position
            style (reportlab.lib.styles.StyleSheet1): optional stylesheet
        '''
        if not style:
            style = self.styles["Normal"]
        p = Paragraph(text, style=style)
        p.wrapOn(self.canvas, self.width, self.height)
        p.drawOn(self.canvas, *self.coord(x, y, mm))


from reportlab.lib.colors import HexColor
import pathlib
from os import path

DEFAULT_SPACING = 8
DEFAULT_STARTING_POSITION = 24
CIRCLE_SIZE = 4

DEFAULT_FONT = 'Helvetica'
DEFAULT_FONT_BOLD = 'Helvetica-Bold'

DEFAULT_LOGO_PATH = path.join(pathlib.Path(__file__).parent.resolve(), 'assets', 'frigel_logo.png')

WORKING_MODE_TYPES = ['standard', 'production', 'maintenance']

DEFAULT_ON_COLOR = HexColor("#39D263")
DEFAULT_OFF_COLOR = HexColor("#E9E9E9")
DEFAULT_STROKE_COLOR = HexColor("#E9E9E9")
WORKING_MODES_COLORS = (HexColor("#6bce70"), HexColor("#bedaa4"), HexColor("#4e70ab"))


from pyfrigel_report_tool.consts import DEFAULT_STROKE_COLOR, WORKING_MODES_COLORS, DEFAULT_FONT

from reportlab.graphics.shapes import Drawing, _DrawingEditorMixin
from reportlab.graphics.charts.legends import Legend
from reportlab.graphics.charts.piecharts import Pie
from reportlab.graphics.charts.barcharts import VerticalBarChart
from reportlab.lib.colors import PCMYKColor
from reportlab.lib.formatters import DecimalFormatter


class ChartPieWithLegend(_DrawingEditorMixin, Drawing):
    '''
        creates a pie chart with a legend
    '''
    def __init__(self, data: list, seriesNames: list, colors: list=list(), width: int=270, height: int=100, *args, **kw):
        Drawing.__init__(self, width, height, *args, **kw)
        self._add(self, Pie(), name='pie', validate=None, desc=None)
        self.pie.data = data
        if colors:
            self._colors = colors
        self.pie.height = 100
        self.pie.width = 100
        self.pie.x = 0
        self.pie.strokeWidth = 1
        self.pie.slices.strokeColor = DEFAULT_STROKE_COLOR
        self.pie.slices.strokeWidth = 0
        self._seriesNames = seriesNames
        self._add(self, Legend(), name='legend', validate=None, desc=None)
        self.legend.y = self.height/2
        self.legend.columnMaximum  = 6
        self.legend.y = self.height/2 + 10
        self.legend.x = 120
        self.legend.fontName = DEFAULT_FONT
        self.legend.strokeWidth = 0
        self.legend.strokeColor = DEFAULT_STROKE_COLOR
        self.legend.fontSize = 14
        self.legend.alignment='right'
        self.legend.subCols.rpad = 40
        for i, _ in enumerate(self.pie.data): self.pie.slices[i].fillColor = self._colors[i]
        self.legend.colorNamePairs = list(zip(self._colors, list(zip(self._seriesNames, map(lambda x: "{}{}".format(x, '%'), self.pie.data)))))
        
        
class ChartBar(_DrawingEditorMixin,Drawing):
    """
        creates a chart bar
        
        input:
            data (list)
            category_names (list): x axys names
            colors (list): bar colors
            width (float)
            height (float)
            
    """
    def __init__(self, data: list, category_names: list, colors: list, width=500, height=200, style='parallel', *args, **kw):
        Drawing.__init__(self,width,height,*args,**kw)
        self._add(self,VerticalBarChart(),name='chart',validate=None,desc=None)
        self.chart.x = 30
        self.chart.y = 20
        self.chart.width=self.width - self.chart.x -10
        self.chart.height=self.height - self.chart.y -10
        self.chart.reversePlotOrder = 0
        
        self.chart.valueAxis.strokeWidth = 0.5
        self.chart.categoryAxis.strokeWidth = 0.5
        self.chart.categoryAxis.style = style
        self.chart.valueAxis.valueMin = 0
        self.chart.valueAxis.valueMax = 24
        self.chart.valueAxis.valueStep= 6
        self.chart.valueAxis.labelTextFormat= '%0.0f h'
        self.chart.data = data
        for index, color in enumerate(colors): self.chart.bars[index].fillColor = color 
        self.chart.bars.strokeWidth = 0.25
        self.chart.bars.strokeColor = DEFAULT_STROKE_COLOR
        self.chart.categoryAxis.categoryNames = category_names
        

class PieChartWorkingModes(_DrawingEditorMixin,Drawing):
    """
        creates a chart bar
        
        input:
            data (list)
            category_names (list): x axys names
            colors (list): bar colors
            width (float)
            height (float)
            
    """
    def __init__(self, data, data_trend, header_names: tuple, series_names: tuple, width=500, height=125, *args, **kw):
        Drawing.__init__(self,width,height,*args,**kw)
        self._colors = WORKING_MODES_COLORS
        # font
        fontSize = 12
        fontName = DEFAULT_FONT
        self._add(self, Pie(), name='chart', validate=None, desc=None)
        # pie
        self.chart.y = 0
        self.chart.x = 0
        self.chart.height = 100
        self.chart.width = 100
        self.chart.slices.strokeColor = DEFAULT_STROKE_COLOR
        self.chart.slices.strokeWidth = 0.5
        self._add(self, Legend(), name='legend', validate=None, desc=None)
        self._add(self, Legend(), name='legendHeader', validate=None, desc=None)
        self.legendHeader.x = 150
        self.legendHeader.y = self.height - 10
        self.legendHeader.fontSize = fontSize
        self.legendHeader.fontName = fontName
        self.legendHeader.subCols[0].minWidth = 200
        self.legendHeader.subCols[0].align = 'left'
        self.legendHeader.subCols[1].minWidth = 80
        self.legendHeader.subCols[1].align = 'right'
        self.legendHeader.subCols[2].minWidth = 160
        self.legendHeader.subCols[2].align = 'right'
        self.legendHeader.subCols[3].minWidth = 1000 # needed to remove black rectangle
        black = PCMYKColor(0, 0, 0, 100)
        self.legendHeader.colorNamePairs = [(black, header_names + ('', ))]
        self.legend.x = 150
        self.legend.y = 90
        self.legend.fontSize = fontSize
        self.legend.fontName = fontName
        self.legend.dx = 8
        self.legend.dy = 8
        self.legend.dxTextSpace = 10
        self.legend.yGap = 0
        self.legend.deltay = 24
        self.legend.strokeColor = PCMYKColor(0,0,0,0)
        self.legend.strokeWidth = 0
        self.legend.columnMaximum = 99
        self.legend.alignment = 'right'
        self.legend.variColumn = 0
        self.legend.dividerDashArray = None
        self.legend.dividerWidth = 0.5
        self.legend.dividerOffsX = (0, 0)
        self.legend.dividerLines = 7
        self.legend.dividerOffsY = 12
        self.legend.subCols[0].align = 'left'
        self.legend.subCols[0].minWidth = 200
        self.legend.subCols[1].align = 'right'
        self.legend.subCols[1].align='numeric'
        self.legend.subCols[1].dx = -30
        self.legend.subCols[1].minWidth = 80
        self.legend.subCols[2].align = 'right'
        self.legend.subCols[2].align='numeric'
        self.legend.subCols[2].dx = -10
        self.legend.subCols[2].minWidth = 160
        # sample data
        self._seriesNames = series_names
        self._seriesData1 = data
        self._seriesData2 = data_trend
        formatter_time = DecimalFormatter(places=1, thousandSep=',', decimalSep='.', suffix=' h')
        formatter_trend = lambda x: DecimalFormatter(places=0, thousandSep=',', decimalSep='.', suffix=' %').format(x) if x != None else x
        sign_adder = lambda x: '-' if x== None else x if x.startswith('-') else '+{}'.format(x)
        names = list(zip(self._seriesNames,
        map(formatter_time, self._seriesData1),
        map(sign_adder, map(formatter_trend, self._seriesData2))))
        self.legend.colorNamePairs = list(zip(self._colors, names))
        self.chart.data  = self._seriesData1
        # apply colors to slices
        for i, _ in enumerate(self.chart.data): self.chart.slices[i].fillColor = self._colors[i]
        self.legend.deltax = 75
        self.legendHeader.subCols[0].minWidth = 100
        self.legend.subCols[0].minWidth = 90